@@ -15,14 +15,14 @@
 //
 // Copyright (C) 2005 Martin Kretzschmar <martink@gnome.org>
 // Copyright (C) 2005 Kristian Høgsberg <krh@redhat.com>
-// Copyright (C) 2006-2008, 2012, 2013, 2015, 2017, 2018 Albert Astals Cid <aacid@kde.org>
+// Copyright (C) 2006-2008, 2012, 2013, 2015, 2017-2019 Albert Astals Cid <aacid@kde.org>
 // Copyright (C) 2007 Brad Hards <bradh@kde.org>
 // Copyright (C) 2009-2013 Thomas Freitag <Thomas.Freitag@alfa.de>
 // Copyright (C) 2009 Till Kamppeter <till.kamppeter@gmail.com>
 // Copyright (C) 2009 Carlos Garcia Campos <carlosgc@gnome.org>
-// Copyright (C) 2009, 2011, 2015-2017 William Bader <williambader@hotmail.com>
+// Copyright (C) 2009, 2011, 2015-2017, 2020 William Bader <williambader@hotmail.com>
 // Copyright (C) 2010 Hib Eris <hib@hiberis.nl>
-// Copyright (C) 2011, 2014, 2017 Adrian Johnson <ajohnson@redneon.com>
+// Copyright (C) 2011, 2014, 2017, 2020 Adrian Johnson <ajohnson@redneon.com>
 // Copyright (C) 2012 Fabio D'Urso <fabiodurso@hotmail.it>
 // Copyright (C) 2018 Klarälvdalens Datakonsult AB, a KDAB Group company, <info@kdab.com>. Work sponsored by the LiMux project of the city of Munich
 // Copyright (C) 2018 Adam Reichold <adam.reichold@t-online.de>
@@ -38,7 +38,7 @@
 #define PSOUTPUTDEV_H
 
 #include "poppler-config.h"
-#include <stddef.h>
+#include <cstddef>
 #include "Object.h"
 #include "GfxState.h"
 #include "GlobalParams.h"
@@ -87,6 +87,12 @@ enum PSOutCustomCodeLocation {
   psOutCustomPageSetup
 };
 
+enum PSForceRasterize {
+  psRasterizeWhenNeeded,	// default
+  psAlwaysRasterize,		// always rasterize, useful for testing
+  psNeverRasterize		// never rasterize, may produce incorrect output
+};
+
 typedef void (*PSOutputFunc)(void *stream, const char *data, int len);
 
 typedef GooString *(*PSOutCustomCodeCbk)(PSOutputDev *psOut,
@@ -106,7 +112,7 @@ public:
 	      bool duplexA = true,
 	      int imgLLXA = 0, int imgLLYA = 0,
 	      int imgURXA = 0, int imgURYA = 0,
-	      bool forceRasterizeA = false,
+	      PSForceRasterize forceRasterizeA = psRasterizeWhenNeeded,
 	      bool manualCtrlA = false,
 	      PSOutCustomCodeCbk customCodeCbkA = nullptr,
 	      void *customCodeCbkDataA = nullptr);
@@ -122,13 +128,13 @@ public:
 	      bool duplexA = true,
 	      int imgLLXA = 0, int imgLLYA = 0,
 	      int imgURXA = 0, int imgURYA = 0,
-	      bool forceRasterizeA = false,
+	      PSForceRasterize forceRasterizeA = psRasterizeWhenNeeded,
 	      bool manualCtrlA = false,
 	      PSOutCustomCodeCbk customCodeCbkA = nullptr,
 	      void *customCodeCbkDataA = nullptr);
 
   // Destructor -- writes the trailer and closes the file.
-  virtual ~PSOutputDev();
+  ~PSOutputDev() override;
 
   // Check if file was successfully created.
   virtual bool isOk() { return ok; }
@@ -166,9 +172,9 @@ public:
   //----- header/trailer (used only if manualCtrl is true)
 
   // Write the document-level header.
-  void writeHeader(const std::vector<int> &pages,
+  void writeHeader(int nPages,
 		   const PDFRectangle *mediaBox, const PDFRectangle *cropBox,
-		   int pageRotate, char *pstitle);
+		   int pageRotate, const char *title);
 
   // Write the Xpdf procset.
   void writeXpdfProcset();
@@ -271,7 +277,7 @@ public:
   void unsetSoftMaskFromImageMask(GfxState *state, double *baseMatrix) override;
   void drawImage(GfxState *state, Object *ref, Stream *str,
 			 int width, int height, GfxImageColorMap *colorMap,
-			 bool interpolate, int *maskColors, bool inlineImg) override;
+			 bool interpolate, const int *maskColors, bool inlineImg) override;
   void drawMaskedImage(GfxState *state, Object *ref, Stream *str,
 			       int width, int height,
 			       GfxImageColorMap *colorMap,
@@ -315,6 +321,7 @@ public:
 
   void setPSCenter(bool center) { psCenter = center; }
   void setRasterAntialias(bool a) { rasterAntialias = a; }
+  void setForceRasterize(PSForceRasterize f) { forceRasterize = f; }
   void setRasterResolution(double r) { rasterResolution = r; }
   void setRasterMono(bool b) { rasterMono = b; }
   void setUncompressPreloadedImages(bool b) { uncompressPreloadedImages = b; }
@@ -383,29 +390,29 @@ private:
   void setupForm(Ref id, Object *strObj);
   void addProcessColor(double c, double m, double y, double k);
   void addCustomColor(GfxSeparationColorSpace *sepCS);
-  void doPath(GfxPath *path);
+  void doPath(const GfxPath *path);
   void maskToClippingPath(Stream *maskStr, int maskWidth, int maskHeight, bool maskInvert);
   void doImageL1(Object *ref, GfxImageColorMap *colorMap,
 		 bool invert, bool inlineImg,
 		 Stream *str, int width, int height, int len,
-		 int *maskColors, Stream *maskStr,
+		 const int *maskColors, Stream *maskStr,
 		 int maskWidth, int maskHeight, bool maskInvert);
   void doImageL1Sep(Object *ref, GfxImageColorMap *colorMap,
 		    bool invert, bool inlineImg,
 		    Stream *str, int width, int height, int len,
-		    int *maskColors, Stream *maskStr,
+		    const int *maskColors, Stream *maskStr,
 		    int maskWidth, int maskHeight, bool maskInvert);
-  void doImageL2(Object *ref, GfxImageColorMap *colorMap,
+  void doImageL2(GfxState *state, Object *ref, GfxImageColorMap *colorMap,
 		 bool invert, bool inlineImg,
 		 Stream *str, int width, int height, int len,
-		 int *maskColors, Stream *maskStr,
+		 const int *maskColors, Stream *maskStr,
 		 int maskWidth, int maskHeight, bool maskInvert);
-  void doImageL3(Object *ref, GfxImageColorMap *colorMap,
+  void doImageL3(GfxState *state, Object *ref, GfxImageColorMap *colorMap,
 		 bool invert, bool inlineImg,
 		 Stream *str, int width, int height, int len,
-		 int *maskColors, Stream *maskStr,
+		 const int *maskColors, Stream *maskStr,
 		 int maskWidth, int maskHeight, bool maskInvert);
-  void dumpColorSpaceL2(GfxColorSpace *colorSpace,
+  void dumpColorSpaceL2(GfxState *state, GfxColorSpace *colorSpace,
 			bool genXform, bool updateColors,
 			bool map01);
   bool tilingPatternFillL1(GfxState *state, Catalog *cat, Object *str,
@@ -429,13 +436,13 @@ private:
   GooString *filterPSName(const GooString *name);
 
   // Write the document-level setup.
-  void writeDocSetup(PDFDoc *doc, Catalog *catalog, const std::vector<int> &pages, bool duplexA);
+  void writeDocSetup(Catalog *catalog, const std::vector<int> &pageList, bool duplexA);
 
   void writePSChar(char c);
   void writePS(const char *s);
   void writePSBuf(const char *s, int len);
   void writePSFmt(const char *fmt, ...);
-  void writePSString(const GooString *s);
+  void writePSString(const std::string &s);
   void writePSName(const char *s);
   GooString *filterPSLabel(GooString *label, bool *needParens=nullptr);
   void writePSTextLine(const GooString *s);
@@ -479,7 +486,7 @@ private:
   int fontIDSize;		// size of fontIDs array
   std::set<int> resourceIDs;	// list of object IDs of objects containing Resources we've already set up
   std::unordered_set<std::string> fontNames; // all used font names
-  std::unordered_map<std::string, int> fontMaxValidGlyph; // max valid glyph of each font
+  std::unordered_map<std::string, int> perFontMaxValidGlyph; // max valid glyph of each font
   PST1FontName *t1FontNames;	// font names for Type 1/1C fonts
   int t1FontNameLen;		// number of entries in t1FontNames array
   int t1FontNameSize;		// size of t1FontNames array
@@ -530,7 +537,7 @@ private:
   bool t3FillColorOnly;	// operators should only use the fill color
   bool t3Cacheable;		// cleared if char is not cacheable
   bool t3NeedsRestore;		// set if a 'q' operator was issued
-  bool forceRasterize;		// forces the page to be rasterized into a image before printing
+  PSForceRasterize forceRasterize; // controls the rasterization of pages into images
   bool displayText;		// displayText
   bool psCenter;		// center pages on the paper
   bool rasterAntialias;	// antialias on rasterize
@@ -557,6 +564,8 @@ private:
   bool enableLZW;		// enable LZW compression
   bool enableFlate;		// enable Flate compression
 
+  std::unordered_set<std::string> iccEmitted; // contains ICCBased CSAs that have been emitted
+
 #ifdef OPI_SUPPORT
   int opi13Nest;		// nesting level of OPI 1.3 objects
   int opi20Nest;		// nesting level of OPI 2.0 objects
